/*
 * PIDController.hpp
 *
 *  Created on: 27 Feb 2018
 *      Author: m2
 */

#ifndef OMNI_CONTROL_M2_CTLRS_INCLUDE_PIDFCONTROLLER_HPP_
#define OMNI_CONTROL_M2_CTLRS_INCLUDE_PIDFCONTROLLER_HPP_

#include <ScalarController.hpp>
#include <cmath>
#include <ros/ros.h>
#include <m2_ctlrs/PIDFSet.h>
#include <geometry_msgs/Twist.h>
#include <geometry_msgs/TwistWithCovarianceStamped.h>

class PIDFController : public ScalarController {
public:
	// Constructor
	PIDFController(double kP, double kI, double kD, double kFF, double iLimit, double outLimit, double deadzone);


	// Output is 0 if abs(input-feedback) < deadzone
	double deadzone;

	// The meat
	void calculate();
protected:
	// Implemented method
	void clearError();

	// Coefficients
	double kP, kI, kD, kFF;

	// Output is clamped between (-output_limit, output_limit)
	double integral_limit;
	double output_limit;
private:
	// Internal variables used in calculate()
	double error_old;
	double error_I;
	double error_D;
};

PIDFController::PIDFController(double kP=1, double kI=0, double kD=0, double kFF=0
		, double iLimit=10, double outLimit=10, double deadzone=0) : kP(kP),kI(kI),kD(kD),kFF(kFF)
,integral_limit(iLimit),output_limit(outLimit),deadzone(deadzone) {
	clearError();
}

void PIDFController::calculate() {
	/*
	 * Do the PID controller calculation , pass in error and return output
	 * Two extra functionalities:
	 * - Limiting the integral term
	 * - Low pass filter for the derivative controller
	 */
//	ROS_INFO("P %f I %f D %f FF %f", kP, kI, kD, kFF);
	double error = input - feedback;
	error_D = error - error_old;

	if (std::fabs(error) > deadzone) {
//		ROS_INFO("err out deadzone");
		error_I += error;
		if (std::fabs(error_I) > integral_limit) {
			error_I = (error_I > 0 ? 1: -1) * integral_limit; //keep sign of error_I
		}
	} else {
//		ROS_INFO("err in deadzone");
		error = 0;
		error_old = 0;
		// error_I = 0; // keeping the steady state error
		error_D = 0;
	}

	// Decay a bit of error
	error_I *= 0.99;

	output = error*kP + error_I*kI + error_D*kD + input*kFF;
//	ROS_INFO("i %f f %f e %f o %f", input, feedback, error, output);
	if (output > output_limit) output = output_limit;
	if (output < -output_limit) output = -output_limit;
}

void PIDFController::clearError() {
	error_old = 0;
	error_I = 0;
	error_D = 0;
}

/**
 * PID controller operating on 64-bit floats
 */

class PIDCtlrROS : public PIDFController {
public:
	PIDCtlrROS(ros::NodeHandle& nh, ros::NodeHandle& nh_priv, std::string prefix
			, double kP=1, double kI=0, double kD=0, double kFF=0
			, double iLimit=10, double outLimit=10, double deadzone=0.01)
:PIDFController(kP,kI,kD,kFF,iLimit,outLimit,deadzone) {

		// DO NOT USE SERVICE, IT SEGFAULTS AND I DONT WANT TO DEBUG NOW
		pidset_srv = nh.advertiseService(prefix+std::string("/pid_set"),&PIDCtlrROS::pidset_srv_cb,this);
		// Get PIDF param for each direction, from nh_priv

		// Integral limit and output limit is set here, as these are not usually required to change
		// Deadzone is also the same. Change them here

		nh_priv.param(prefix+std::string("/kP"),this->kP,0.0);
		nh_priv.param(prefix+std::string("/kI"),this->kI,0.0);
		nh_priv.param(prefix+std::string("/kD"),this->kD,0.0);
		nh_priv.param(prefix+std::string("/kFF"),this->kFF,1.0);
		nh_priv.param(prefix+std::string("/dz"),this->deadzone,0.01);
		nh_priv.param(prefix+std::string("/iLimit"),this->integral_limit, 100.0);
		nh_priv.param(prefix+std::string("/outLimit"),this->output_limit, 10.0);
	}
private:
	ros::ServiceServer pidset_srv;
	bool pidset_srv_cb(m2_ctlrs::PIDFSet::Request& req, m2_ctlrs::PIDFSet::Response& resp) {
		ROS_INFO("called pidset");
		this->kP = req.param.kP;
		this->kI = req.param.kI;
		this->kD = req.param.kD;
		this->kFF = req.param.kFF;
		resp.stamp = ros::Time::now();
		ROS_INFO("done pidset");
		return true;
	}
};

/**
 * Twist controller working with linear/x, linear/y and angular/z; with ROS integration
 */
class TwistCtlr2D {
public:
	geometry_msgs::Twist input;
	geometry_msgs::Twist feedback;
	geometry_msgs::Twist output;
	TwistCtlr2D(ros::NodeHandle& nh, ros::NodeHandle& nh_priv) : input_is_updated(false), fb_is_updated(false) {
		controllers.push_back(PIDCtlrROS(nh, nh_priv, "x"));
		controllers.push_back(PIDCtlrROS(nh, nh_priv, "y"));
		controllers.push_back(PIDCtlrROS(nh, nh_priv, "z"));


		// Name of the topic of Twist for controlling
		std::string input_topic, output_topic, fb_topic, err_topic;
		nh_priv.param("input", input_topic, std::string("pid_input"));
		nh_priv.param("output", output_topic, std::string("pid_output"));
		nh_priv.param("feedback", fb_topic, std::string("pid_feedback"));
		nh_priv.param("err", err_topic, std::string("pid_err"));


		// Input subscriber and output publisher
		in_sub = nh.subscribe(input_topic,5,&TwistCtlr2D::input_cb,this);
		if (!in_sub) ROS_ERROR("cant make in_sub");
		fb_sub = nh.subscribe(fb_topic,5,&TwistCtlr2D::fb_cb,this);
		if (!fb_sub) ROS_ERROR("cant make fb_sub");
		out_pub = nh.advertise<geometry_msgs::Twist>(output_topic,5);
		err_pub = nh.advertise<geometry_msgs::Vector3>(err_topic,5);

		ROS_INFO("Made TwistCtlr2D");
	}

	void calculate() {
		// Update input
		controllers[0].input = input_buffer.x;
		controllers[1].input = input_buffer.y;
		controllers[2].input = input_buffer.z;

		controllers[0].feedback = fb_buffer.x;
		controllers[1].feedback = fb_buffer.y;
		controllers[2].feedback = fb_buffer.z;

		// Publish err
		geometry_msgs::Vector3 err_msg;
		err_msg.x = input_buffer.x - fb_buffer.x;
		err_msg.y = input_buffer.y - fb_buffer.y;
		err_msg.z = input_buffer.z - fb_buffer.z;
		err_pub.publish(err_msg);

		// Update
		controllers[0].calculate();
		controllers[1].calculate();
		controllers[2].calculate();

		// Output
		geometry_msgs::Twist out_msg;
		out_msg.linear.x = controllers[0].output;
		out_msg.linear.y = controllers[1].output;
		out_msg.angular.z = controllers[2].output;
//		ROS_INFO("x %f y %f z %f", out_msg.linear.x, out_msg.linear.y, out_msg.angular.z);
		out_pub.publish(out_msg);
	}

	void input_cb(const geometry_msgs::Twist::ConstPtr& msg) {
		input_buffer.x = msg->linear.x;
		input_buffer.y = msg->linear.y;
		input_buffer.z = msg->angular.z;
		input_is_updated = true;
	}
	void fb_cb(const geometry_msgs::TwistWithCovarianceStamped::ConstPtr& msg) {
		fb_buffer.x = msg->twist.twist.linear.x;
		fb_buffer.y = msg->twist.twist.linear.y;
		fb_buffer.z = msg->twist.twist.angular.z;
		fb_is_updated = true;
	}
	// Publisher and subscribers
	ros::Subscriber in_sub, fb_sub;
	ros::Publisher out_pub, err_pub;
private:
	std::vector<PIDCtlrROS> controllers;



	// Input data update
	geometry_msgs::Vector3 input_buffer;
	bool input_is_updated;



	// Feedback data update
	geometry_msgs::Vector3 fb_buffer;
	bool fb_is_updated;


};

#endif /* OMNI_CONTROL_M2_CTLRS_INCLUDE_PIDFCONTROLLER_HPP_ */
